package net.knarcraft.serverlauncher;

import net.knarcraft.serverlauncher.profile.Collection;
import net.knarcraft.serverlauncher.profile.Profile;
import net.knarcraft.serverlauncher.server.Server;
import net.knarcraft.serverlauncher.server.ServerType;
import net.knarcraft.serverlauncher.userinterface.GUI;
import net.knarcraft.serverlauncher.userinterface.ServerConsoles;

import javax.naming.ConfigurationException;
import java.awt.*;
import java.io.IOException;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;
//Java 8 required.

/**
 * A software for managing Minecraft servers.
 *
 * @author		Kristian Knarvik <kristian.knarvik@knett.no>
 * @version	0.0.0.1
 * @since		0.0.0.1
 */

public class Main {
    private static GUI gui;

	public static void main(String[] args) {
		EventQueue.invokeLater(() -> {
			try {
				setup();
				gui = new GUI();
				new ServerConsoles();
                Profile.load();
				//TODO: replace with profiles loading generating a default profile if empty.

				ScheduledExecutorService exec = Executors.newSingleThreadScheduledExecutor();
				exec.scheduleAtFixedRate(Main::updateConsoles, 10, 250, TimeUnit.MILLISECONDS);
			} catch (Exception e) {
				e.printStackTrace();
			}
		});
	}

	public static GUI gui() {
	    return gui;
    }

	private static void setup() {
		try {
			ServerType.loadServerTypes();
		} catch (ConfigurationException e) {
			e.printStackTrace();
			System.exit(1);
		}
	}

    /**
     * Reads from server processes, and writes the output to our custom consoles.
     */
	private static void updateConsoles() {
        for (Collection collection : Profile.getCurrent().getCollections()) {
            Server server = collection.getServer();
            if (server.isEnabled() && server.getProcess() != null) {
                try {
                    String readText = server.read();
                    if (!readText.equals("")) {
                        collection.getServerConsole().output(readText);
                        updatePlayerList(readText, collection);
                    }
                } catch (IOException e) {
                    e.printStackTrace();
                }
            }
        }
    }

    /**
     * Looks for strings implying a player has joined or left, and updates the appropriate lists.
     *
     * @param text          The text to search.
     * @param collection    The server which sent the text.
     */
	private static void updatePlayerList(String text, Collection collection) {
	    Server server = collection.getServer();
	    if (!server.typeName().equals("Bungee")) {
            String joinedPlayer = stringBetween(text, "[Server thread/INFO]: ", " joined the game");
            if (!joinedPlayer.equals("")) {
                if (!server.hasPlayer(joinedPlayer)) {
                    server.addPlayer(joinedPlayer);
                }
            } else {
                joinedPlayer = stringBetween(text, "UUID of player ", " is ");
                if (!joinedPlayer.equals("")) {
                    if (!server.hasPlayer(joinedPlayer)) {
                        server.addPlayer(joinedPlayer);
                    }
                }
            }
            if (joinedPlayer.equals("")) {
                String leftPlayer = stringBetween(text, "INFO]: ", " lost connection");
                if (!leftPlayer.equals("")) {
                    if (server.hasPlayer(leftPlayer)) {
                        server.removePlayer(leftPlayer);
                    }
                } else {
                    leftPlayer = stringBetween(text, "[Server thread/INFO]: ", " left the game");
                    if (!leftPlayer.equals("")) {
                        if (server.hasPlayer(leftPlayer)) {
                            server.removePlayer(leftPlayer);
                        }
                    }
                }
            }
        }
    }

    /**
     * Finds a substring between two substrings in a string.
     *
     * @param string The string containing the substrings
     * @param start  The substring before the wanted substring
     * @param end    The substring after the wanted substring
     * @return The wanted substring.
     */
    private static String stringBetween(String string, String start, String end) {
        int startPos = string.indexOf(start) + start.length();
        if (!string.contains(start) || string.indexOf(end, startPos) < startPos) {
            return "";
        }
        return string.substring(startPos, string.indexOf(end, startPos));
    }
}